#include "sdlskk_internal.h"
#include <ctype.h>
#include <assert.h>

#define INPUT_BUF_SIZE 2048
#define HENKAN_BUF_SIZE 512
#define OKURIGANA_BUF_SIZE 10
#define PREFIX_BUF_SIZE 10

#define START_MINIBUFFER_WORDSELECT 4
#define MINIBUFFER_WORDSELECT_MAX_WORDS 7

SDLSKK_Context* SDLSKK_Context_new( SDLSKK_Dictionary* dict,
				    SDLSKK_RomKanaRuleTable* table,
				    SDLSKK_Keybind* keybind,
				    int use_minibuffer )
{
  SDLSKK_Context* context = SDLSKK_malloc( sizeof(SDLSKK_Context) );
  
  context->dict = dict;
  context->rule_table = table;
  context->mode = SDLSKK_J_KAKUTEI_MODE;
  context->str = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->prefix = SDLSKK_XStr_new4( "", PREFIX_BUF_SIZE );
  context->henkan_key = SDLSKK_XStr_new4 ( "", HENKAN_BUF_SIZE );
  context->henkan_okurigana = SDLSKK_XStr_new4( "", OKURIGANA_BUF_SIZE );
  context->dict_item = NULL;
  context->display_str = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->minibuffer_str = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->cursor_pos = 0;
  context->henkan_buf_pos = 0;
  context->katakana_on = 0;
  context->use_minibuffer = use_minibuffer;
  context->child_context = NULL;
  context->parent_context = NULL;
  context->cut_buffer = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->marked_point = NOT_MARKED;
  context->keybind = keybind;
  return context;
}

void SDLSKK_Context_delete( SDLSKK_Context* context )
{
  if( context == NULL )
    return;
  SDLSKK_XStr_delete( context->str );
  SDLSKK_XStr_delete( context->prefix );
  SDLSKK_XStr_delete( context->henkan_key );
  SDLSKK_XStr_delete( context->henkan_okurigana );
  SDLSKK_XStr_delete( context->display_str );
  SDLSKK_XStr_delete( context->minibuffer_str );
  SDLSKK_DictItem_delete( context->dict_item );
  SDLSKK_Context_delete( context->child_context );
  SDLSKK_XStr_delete( context->cut_buffer );
  free( context );
}

void SDLSKK_Context_insert_Char( SDLSKK_Context* context, SDLSKK_Char c )
{
  SDLSKK_XStr_insert_Char( context->str, context->cursor_pos, c);
  ++(context->cursor_pos);
}

static void Context_insert_XStr( SDLSKK_Context* context, SDLSKK_XStr* str )
{
  SDLSKK_XStr_insert_Str( context->str, context->cursor_pos, str->buf);
  context->cursor_pos += str->len;
}

void SDLSKK_Context_insert_Str( SDLSKK_Context* context, SDLSKK_Char *str )
{
  SDLSKK_XStr_insert_Str( context->str, context->cursor_pos, str );
  context->cursor_pos += SDLSKK_Str_len( str );
}

static void SDLSKK_XStr_chop( SDLSKK_XStr* str )
{
  SDLSKK_XStr_delete_Char( str, str->len - 1 );
}

static void Context_henkan_kakutei( SDLSKK_Context* context )
{  
  SDLSKK_Context_insert_Str( context,
			     SDLSKK_DictItem_get_current( context->dict_item ) );
  SDLSKK_Context_insert_Str( context, context->henkan_okurigana->buf );
  SDLSKK_DictItem_commit( context->dict_item );
  SDLSKK_DictItem_delete( context->dict_item );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->dict_item = NULL;
  context->mode = SDLSKK_J_KAKUTEI_MODE;
    
}

static void Context_unset_mark( SDLSKK_Context* context )
{
  context->marked_point = NOT_MARKED;
}

static SDLSKK_Char* convert_hiragana_katakana( SDLSKK_Char* str, int katanana )
{
  if( katanana )
    return SDLSKK_Str_hiragana_to_katakana( str );
  else
    return SDLSKK_Str_katakana_to_hiragana( str );
}

static SDLSKK_Char* get_kana( SDLSKK_RomKanaElement* kana, int katakana_on )
{
  return katakana_on?kana->katakana:kana->hiragana;
}

static void resolve_romkana( SDLSKK_RomKanaRuleTable* table,
			     SDLSKK_XStr* prefix, SDLSKK_XStr* inserted_str,
			     int* pos, int katakana_on )
{
  SDLSKK_RomKanaElement* kana;
  SDLSKK_XStr* prefix_buf;
  
  if( prefix->len == 0 )
    return;
  
  switch( SDLSKK_RomKanaRuleTable_get( table, prefix->buf, &kana ) ){
    
  case SDLSKK_RomKana_NoCandidate:
    prefix_buf = SDLSKK_XStr_new5( prefix );
    SDLSKK_XStr_chop( prefix_buf );
    
    if( SDLSKK_RomKanaRuleTable_get( table, prefix_buf->buf, &kana ) == 
	SDLSKK_RomKana_UnresolvedCandidate ){
      SDLSKK_Char rest;
      SDLSKK_XStr_insert_Str( inserted_str, *pos,
			      get_kana( kana, katakana_on ) );
      *pos += SDLSKK_Str_len( get_kana( kana, katakana_on ) );
            
     
      rest = prefix->buf[prefix->len-1];
      SDLSKK_XStr_clear( prefix );
      SDLSKK_XStr_insert_Char( prefix, 0, rest );
      
      resolve_romkana( table, prefix, inserted_str, pos, katakana_on );
    }else{
      SDLSKK_XStr_delete_Char( prefix, 0 );
      resolve_romkana( table, prefix, inserted_str, pos, katakana_on );
    }
    SDLSKK_XStr_delete( prefix_buf );
    return;
    
  case SDLSKK_RomKana_SomeCandidates:
    return;
    
  case SDLSKK_RomKana_OneCandidate:
    SDLSKK_XStr_insert_Str( inserted_str, *pos, get_kana( kana, katakana_on ));
    *pos += SDLSKK_Str_len( get_kana( kana, katakana_on ) );

    SDLSKK_XStr_copy_Str( prefix, kana->next_state );
    return;
    
  case SDLSKK_RomKana_UnresolvedCandidate:
    return;
  }
}

static void resolve_finish( SDLSKK_RomKanaRuleTable* rule_table,
			    SDLSKK_XStr* prefix,
			    SDLSKK_XStr* inserted_str,
			    int* pos )
{
  SDLSKK_RomKanaElement* candidate;
  if( SDLSKK_RomKanaRuleTable_get( rule_table, prefix->buf, &candidate ) ==
      SDLSKK_RomKana_UnresolvedCandidate ){
    SDLSKK_XStr_insert_Str( inserted_str, *pos, candidate->hiragana );
    *pos += SDLSKK_Str_len( candidate->hiragana );
    SDLSKK_XStr_clear( prefix );
  }else{
    SDLSKK_XStr_clear( prefix );
  }
}

static int Context_is_in_word_register_mode( SDLSKK_Context* context )
{
  return ( context->parent_context != NULL );
}

static int Context_is_in_minibuffer_wordselect_mode( SDLSKK_Context* context )
{
  if( context->dict_item == NULL )
    return 0;
  if( ! context->use_minibuffer )
    return 0;
  return ( SDLSKK_DictItem_get_count( context->dict_item ) >=
	   START_MINIBUFFER_WORDSELECT ) ;
}

static void normal_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len , key );
  resolve_romkana( context->rule_table, context->prefix, context->str,
		   &context->cursor_pos, context->katakana_on  );
}
static void upper_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->henkan_buf_pos = 0;
  context->mode = SDLSKK_J_HENKAN_ON_MODE;
  
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len,
			   tolower( key ) );      
  resolve_romkana( context->rule_table, context->prefix,
		   context->henkan_key, &(context->henkan_buf_pos),
		   0 );
}  

static void graph_on_latin_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  SDLSKK_Context_insert_Char( context, key );
}
static void change_to_latin_mode_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_LATIN_MODE;
  SDLSKK_XStr_clear( context->prefix );
  return;
}
static void change_to_jisx0208_latin_mode_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_JISX0208_LATIN_MODE;
  SDLSKK_XStr_clear( context->prefix );
  return;
}

static void change_to_abbrev_mode_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  context->mode = SDLSKK_ABBREV_MODE;
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  context->henkan_buf_pos = 0;      
  return;
}

static void graph_on_j_henkan_okuri_mode( SDLSKK_Context* context, char key )
{
  int pos;

  key = tolower(key);
  
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len, key );
  pos = context->henkan_okurigana->len;
  resolve_romkana( context->rule_table, context->prefix,
		   context->henkan_okurigana, &pos, 0 );
  if( context->prefix->len == 0 ){
    SDLSKK_XStr_insert_Char( context->henkan_key, context->henkan_key->len,
			     context->okuri_char );
    context->dict_item = SDLSKK_Dict_get( context->dict,
					  context->henkan_key->buf );
    SDLSKK_XStr_chop( context->henkan_key );
    if( context->dict_item == NULL ){
      context->mode = SDLSKK_J_HENKAN_ON_MODE;
      SDLSKK_XStr_clear( context->henkan_okurigana );
      return;
    }else{
      context->mode = SDLSKK_J_HENKAN_ACTIVE_MODE;
      return;
    }
  }
}

static void normal_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len, key );
  resolve_romkana( context->rule_table, context->prefix,
		   context->henkan_key, &(context->henkan_buf_pos),
		   0 );
}

static void upper_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  if( context->henkan_key->len == 0 ){
    normal_on_j_henkan_on_mode( context, tolower(key) );
    return;
  }
  if( context->prefix->len != 0 ){
    normal_on_j_henkan_on_mode( context, tolower(key) );
    if( context->prefix->len == 0 ){      
      return;
    }else{
      context->okuri_char = tolower( key );
      SDLSKK_XStr_clear( context->henkan_okurigana );
      context->mode = SDLSKK_J_HENKAN_OKURI_MODE;
      return;
    }  
  }
  context->okuri_char = tolower( key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->mode = SDLSKK_J_HENKAN_OKURI_MODE;
  SDLSKK_Context_input_key( context, tolower( key ) );
}

static void change_to_latin_mode_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  Context_insert_XStr( context, context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_LATIN_MODE;
}
static void change_to_jisx0208_latin_mode_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  Context_insert_XStr( context, context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_JISX0208_LATIN_MODE;
}

static void graph_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  if( islower(key) ){
    normal_on_j_henkan_on_mode( context, key );
  }
  if( isupper(key) ){
    upper_on_j_henkan_on_mode( context, key );
  }
}

static void graph_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  /*
    generated by
     ruby -e ' a = Array.new(256,-1); a[?a]=0; a[?s]=1; a[?d]=2; a[?f]=3; \
     a[?j]=4; a[?k]=5; a[?l]=6 ; p a' > ary
  */
  int i;
  SDLSKK_DictionaryItem item;
  static const int cnt[] = {
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, -1, -1, 2, -1, 3, -1, -1, -1, 4, 5, 6, -1, -1, -1, -1, -1, -1, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  };
  
  if( Context_is_in_minibuffer_wordselect_mode( context ) ){
    if( cnt[key]==-1 )
      return;
    item = *(context->dict_item);
    for( i=0; i < cnt[key]; ++i ){
      if( SDLSKK_DictItem_is_end( &item ) )
	return;
      SDLSKK_DictItem_next( &item );
    }
    *(context->dict_item) = item;
    Context_henkan_kakutei( context );
    return;
  }
  Context_henkan_kakutei( context );
  SDLSKK_Context_input_key( context, key );
}

static void graph_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_insert_Char( context->henkan_key, context->henkan_buf_pos, key );
  ++(context->henkan_buf_pos);
}

static void graph_on_jisx0208_latin_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  SDLSKK_Context_insert_Char( context, SDLSKK_latin_to_jisx0208_Char( key ) );
}

static void toggle_katakana_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  context->katakana_on = !context->katakana_on;
}

static void toggle_kata_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Char* str;

  resolve_finish( context->rule_table, context->prefix, context->henkan_key,
		  &context->henkan_buf_pos );
  str = convert_hiragana_katakana( context->henkan_key->buf,
				   ! context->katakana_on );
  SDLSKK_Context_insert_Str( context, str );
  free( str );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void space_on_latin_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Context_insert_Char( context, ' ' );
}
static void space_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_Context_insert_Char( context, ' ' );
}


static void change_to_word_register_mode( SDLSKK_Context* context )
{
  context->child_context = SDLSKK_Context_new( context->dict,
					       context->rule_table,
					       context->keybind,
					       0 );
  context->child_context->parent_context = context;
}
static void space_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  resolve_finish( context->rule_table, context->prefix,
		  context->henkan_key, &context->henkan_buf_pos );
  
  context->dict_item = SDLSKK_Dict_get( context->dict, context->henkan_key->buf );
  if( context->dict_item == NULL ){
    change_to_word_register_mode( context );
  }else{
    context->mode = SDLSKK_J_HENKAN_ACTIVE_MODE;
  }
}
static void space_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  int i;
    
  if( Context_is_in_minibuffer_wordselect_mode( context ) ){
    for( i=0; i < MINIBUFFER_WORDSELECT_MAX_WORDS; ++i ){
      if( SDLSKK_DictItem_is_end( context->dict_item ) ){
	change_to_word_register_mode( context );
	return;
      }
      SDLSKK_DictItem_next( context->dict_item );
    }
    return;
  }
  if( SDLSKK_DictItem_is_end( context->dict_item ) ){
    change_to_word_register_mode( context );
    return;
  }
  SDLSKK_DictItem_next( context->dict_item );  
}
static void space_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  context->dict_item = SDLSKK_Dict_get( context->dict, context->henkan_key->buf );
  if( context->dict_item == NULL ){
    change_to_word_register_mode( context );
  }else{
    context->mode = SDLSKK_J_HENKAN_ACTIVE_MODE;
  }
}

static void kakutei_on_latin_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void kakutei_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Char* kana;
  
  SDLSKK_XStr_clear( context->prefix );
  kana = convert_hiragana_katakana( context->henkan_key->buf,
				    context->katakana_on );
  SDLSKK_Context_insert_Str( context, kana );
  free( kana );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
  return;
}
static void kakutei_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  if(  Context_is_in_minibuffer_wordselect_mode( context ) )
    return;
  Context_henkan_kakutei( context );
}

static void kakutei_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Context_insert_Str( context, context->henkan_key->buf );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void kouho_backward_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  int i;
  
  if( Context_is_in_minibuffer_wordselect_mode( context )
      && ( SDLSKK_DictItem_get_count( context->dict_item )
	   != START_MINIBUFFER_WORDSELECT ) ){
    for( i=0; i < MINIBUFFER_WORDSELECT_MAX_WORDS; i++ )
      SDLSKK_DictItem_prev( context->dict_item );
    return;
  }
  
  SDLSKK_DictItem_prev( context->dict_item );
}
static void forward_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->cursor_pos < context->str->len )
    ++(context->cursor_pos);
  return;
}
static void forward_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }
  if( context->cursor_pos < context->str->len )
    ++(context->cursor_pos);
  return;
}


static void backward_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->cursor_pos > 0 )
    --(context->cursor_pos);
  return;
}
static void backward_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }else if( context->cursor_pos > 0 ){
    --(context->cursor_pos);
  }
  return;
}


static void XStr_backspace( SDLSKK_XStr* str, int* pos )
{
  if( *pos <= 0 )
    return;
  SDLSKK_XStr_delete_Char( str, *pos - 1 );
  --(*pos);
  return;
}
static void backspace_on_latin_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  XStr_backspace( context->str, &context->cursor_pos );
  return;
}
static void backspace_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }else{
    XStr_backspace( context->str, &context->cursor_pos );
  }
  return;
}
static void backspace_on_j_henkan_okuri_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->mode = SDLSKK_J_HENKAN_ON_MODE;
}
static void backspace_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }else if( context->henkan_key->len == 0 ){
    context->mode = SDLSKK_J_KAKUTEI_MODE;
  }else{
    XStr_backspace( context->henkan_key, &context->henkan_buf_pos );
  }
}
static void backspace_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  if( Context_is_in_minibuffer_wordselect_mode( context ) )
    return;
  Context_henkan_kakutei( context );
  backspace_on_j_kakutei_mode( context, key );
}
static void backspace_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  if( context->henkan_key->len > 0 ){
    XStr_backspace( context->henkan_key, &context->henkan_buf_pos );
  }else{
    context->mode = SDLSKK_J_KAKUTEI_MODE;
  }
}

static void delete_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->cursor_pos < context->str->len ){
    Context_unset_mark( context );
    context->cursor_pos += 1;
    XStr_backspace( context->str, &context->cursor_pos );
  }
}
static void delete_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  delete_on_latin_mode( context, key );
}

static void cancel_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void cancel_on_j_henkan_okuri_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void cancel_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_DictItem_delete( context->dict_item );
  context->dict_item = NULL;
  context->mode = SDLSKK_J_HENKAN_ON_MODE;
}

static void cancel_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Context* parent_context;
  
  if( ! Context_is_in_word_register_mode( context ) )
    return;
  
  parent_context = context->parent_context;
  parent_context->child_context = NULL;
  SDLSKK_Context_delete( context );
  SDLSKK_DictItem_delete( parent_context->dict_item );
  parent_context->dict_item = NULL;
  parent_context->mode = SDLSKK_J_HENKAN_ON_MODE;
  
}

static void cancel_on_latin_mode( SDLSKK_Context* context, char key )
{
  cancel_on_j_kakutei_mode( context, key );
}

static void cancel_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void register_word_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_DictionaryItem* dict_item;
  SDLSKK_Char* henkan_key;
  
  if( context->parent_context == NULL )
    return;

  if( context->str->len == 0 ){
    cancel_on_j_kakutei_mode( context, key );
    return;
  }
  
  dict_item = context->parent_context->dict_item;
  
  if( dict_item == NULL ){
    henkan_key = SDLSKK_Str_new( context->parent_context->henkan_key->buf );
    SDLSKK_Dict_add_empty_yomi( context->parent_context->dict,
				henkan_key );
    dict_item = SDLSKK_Dict_get( context->parent_context->dict,
				 henkan_key );
    context->parent_context->dict_item = dict_item;
  }
  
  SDLSKK_DictItem_register_word( dict_item, context->str->buf );
  
  SDLSKK_DictItem_set_current( dict_item, context->str->buf );
  Context_henkan_kakutei( context->parent_context );
  
  context->parent_context->child_context = NULL;
  SDLSKK_Context_delete( context );

}

static void yank_on_latin_mode( SDLSKK_Context* context, char key )
{
  Context_unset_mark( context );
  Context_insert_XStr( context, context->cut_buffer );
}

static void yank_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 )
    return;
  yank_on_latin_mode( context, key );
}

static int max2( x, y ){
  return (x>y)?x:y;
}
static int min2( x, y ){
  return (x<y)?x:y;
}

static int Context_begin_region( SDLSKK_Context* context )
{
  return min2( context->marked_point, context->cursor_pos );
}
static int Context_end_region( SDLSKK_Context* context )
{
  return max2( context->marked_point, context->cursor_pos );
}

static void Context_copy_to_buffer( SDLSKK_Context* context )
{
  int begin_region = Context_begin_region( context );
  int end_region = Context_end_region( context );
  
  if( begin_region == end_region ){
    SDLSKK_XStr_clear( context->cut_buffer );
  }else{
    SDLSKK_XStr_slice_Str( context->cut_buffer, context->str->buf, 
			   begin_region, end_region -1 );
  }
}

static void cut_on_latin_mode( SDLSKK_Context* context, char key )
{
  int begin_region = Context_begin_region( context );
  int end_region = Context_end_region( context );
  int i;
  
  if( context->marked_point == NOT_MARKED )
    return;
  
  Context_copy_to_buffer( context );
  Context_unset_mark( context );
  context->cursor_pos = end_region;
  for( i=0; i < end_region - begin_region; ++i )
    backspace_on_latin_mode( context, 0 );
}

static void copy_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->marked_point == NOT_MARKED )
    return;

  Context_copy_to_buffer( context );
}

static void mark_on_latin_mode( SDLSKK_Context* context, char key )
{
  context->marked_point = context->cursor_pos;
}
static void mark_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 )
    return ;
  context->marked_point = context->cursor_pos;
}

static void beginning_of_line_on_latin_mode( SDLSKK_Context* context, char key )
{
  context->cursor_pos = 0;
}

static void beginning_of_line_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }
  context->cursor_pos = 0;
}

static void end_of_line_on_latin_mode( SDLSKK_Context* context, char key )
{
  context->cursor_pos = context->str->len;
}

static void end_of_line_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }
  context->cursor_pos = context->str->len;
}

void SDLSKK_Context_input_event( SDLSKK_Context* context,SDL_Event* event )
{
  if( event->key.keysym.sym < SDLK_FIRST || event->key.keysym.sym > SDLK_LAST )
    return;
  
  if( context->child_context != NULL ){
    SDLSKK_Context_input_event( context->child_context, event );
    return;
  }
  if( event->type == SDL_KEYDOWN ){
    SDLSKK_Command function;
    function = SDLSKK_Keybind_get_command( context->keybind,
					   event->key.keysym, context->mode );
    if( function != NULL )
      function( context, event->key.keysym.unicode );
  }
}


void SDLSKK_Context_input_key( SDLSKK_Context* context, char key )
{
  if( key < 0 )
    return;
  
  if( context->child_context != NULL ){
    SDLSKK_Context_input_key( context->child_context, key );
    return;
  }

 {
   SDLSKK_Command function;
   SDL_keysym keysym = {0,0,0,0};
   if( iscntrl( key ) ){
     keysym.mod = KMOD_CTRL;
     keysym.sym = key + 'a' - 1;
   }
   keysym.unicode = key;
   function = SDLSKK_Keybind_get_command( context->keybind, keysym,
					  context->mode );
   if( function != NULL )
     function( context, key );
 }
  
}

struct CommandStrTable {
  char str[35];
  SDLSKK_Command commands[7];
};

#include "command_str.tbl"

SDLSKK_Command* SDLSKK_convert_command_str( char* command_str )
{
  int i;
  
  for( i=0; strcmp( command_tbl[i].str, "" ) != 0; ++i ){
    if( strcmp( command_tbl[i].str, command_str ) == 0 )
      return command_tbl[i].commands;
  }
  return NULL;
}


static SDLSKK_Char henkan_marker()
{
  static const SDLSKK_Char marker_char[3] = {
    (162<<8)+166,
    0x25BD,
    (129<<8)+164,
  };
  return marker_char[ SDLSKK_get_encoding() ];
}
static SDLSKK_Char henkan_active_marker()     
{
  static const SDLSKK_Char active_marker_char[3] = {
    (162<<8)+167,
    0x25BC,    
    (129<<8)+165,
  };
  return active_marker_char[ SDLSKK_get_encoding() ];
}

static void format_henkan_XStr_on_j_henkan_on_mode( SDLSKK_Context* context, SDLSKK_XStr* xstr )
{
  SDLSKK_Char* henkan_key_katakana;
  
  SDLSKK_XStr_insert_Str( xstr, 0, context->prefix->buf );
  if( context->katakana_on ){
    henkan_key_katakana =
      SDLSKK_Str_hiragana_to_katakana( context->henkan_key->buf );
    SDLSKK_XStr_insert_Str( xstr, 0, henkan_key_katakana );
    free( henkan_key_katakana );
  }else{      
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_key->buf );
  }
  SDLSKK_XStr_insert_Char( xstr, 0, henkan_marker() );
}

static void format_henkan_XStr_on_j_henkan_okuri_mode( SDLSKK_Context* context, SDLSKK_XStr* xstr )
{
  SDLSKK_Char* henkan_key_katakana;
  SDLSKK_Char* henkan_okurigana_katakana;

  SDLSKK_XStr_insert_Str( xstr, 0, context->prefix->buf );
  if( context->katakana_on ){
    henkan_key_katakana =
      SDLSKK_Str_hiragana_to_katakana( context->henkan_key->buf );
    henkan_okurigana_katakana =
      SDLSKK_Str_hiragana_to_katakana( context->henkan_okurigana->buf );
    SDLSKK_XStr_insert_Str( xstr, 0, henkan_okurigana_katakana );
    SDLSKK_XStr_insert_Char( xstr, 0, '*' );
    SDLSKK_XStr_insert_Str( xstr, 0, henkan_key_katakana );
    free( henkan_key_katakana );
    free( henkan_okurigana_katakana );      
  }else{
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_okurigana->buf );
    SDLSKK_XStr_insert_Char( xstr, 0, '*' );
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_key->buf );
  }
  SDLSKK_XStr_insert_Char( xstr, 0, henkan_marker() );
}

void SDLSKK_Context_get_henkan_XStr( SDLSKK_Context* context, SDLSKK_XStr* xstr )
{
  SDLSKK_XStr_clear( xstr );
  
  switch( context->mode ){
  case SDLSKK_LATIN_MODE:
    break;
  case SDLSKK_J_KAKUTEI_MODE:
    SDLSKK_XStr_insert_Str( xstr, 0, context->prefix->buf );
    break;
  case SDLSKK_J_HENKAN_ON_MODE:    
    format_henkan_XStr_on_j_henkan_on_mode( context, xstr );
    break;
  case SDLSKK_J_HENKAN_OKURI_MODE:
    format_henkan_XStr_on_j_henkan_okuri_mode( context, xstr );
    break;
  case SDLSKK_J_HENKAN_ACTIVE_MODE:
    if( ! Context_is_in_minibuffer_wordselect_mode( context ) ){
      SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_okurigana->buf );
      SDLSKK_XStr_insert_Str( xstr, 0, SDLSKK_DictItem_get_current( context->dict_item ) );
    }
    SDLSKK_XStr_insert_Char( xstr, 0, henkan_active_marker() );
    break;
  case SDLSKK_JISX0208_LATIN_MODE:
    break;
  case SDLSKK_ABBREV_MODE:
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_key->buf );
    SDLSKK_XStr_insert_Char( xstr, 0, henkan_marker() );
    break;
  }
}
SDLSKK_Char* SDLSKK_Context_get_display_Str( SDLSKK_Context* context )
{
  SDLSKK_XStr* converting_str;

  if( context->child_context != NULL && !context->use_minibuffer ){
    return SDLSKK_Context_get_display_Str( context->child_context );
  }
  converting_str = SDLSKK_XStr_new4( "", 1024 );
  SDLSKK_XStr_copy_Str( context->display_str, context->str->buf );
  SDLSKK_Context_get_henkan_XStr( context, converting_str );
  SDLSKK_XStr_insert_Str( context->display_str, context->cursor_pos,
			  converting_str->buf );
  SDLSKK_XStr_delete( converting_str );
  return context->display_str->buf;
}

static void XStr_concat_Char( SDLSKK_XStr* str, SDLSKK_Char ch )
{
  SDLSKK_XStr_insert_Char( str, str->len, ch );
}

SDLSKK_Char* SDLSKK_Context_get_minibuffer_Str( SDLSKK_Context* context )
{
  int i;
  SDLSKK_DictionaryItem item;
  static const char asdfjkl[] = { 'A', 'S', 'D', 'F', 'J', 'K', 'L', };
				
  if( ! context->use_minibuffer )
    return NULL;

  if( context->child_context != NULL )
    return SDLSKK_Context_get_display_Str( context->child_context );
  
  if( Context_is_in_minibuffer_wordselect_mode( context ) ){
    item = *(context->dict_item);
    SDLSKK_XStr_clear( context->minibuffer_str );    
    for( i=0; i < MINIBUFFER_WORDSELECT_MAX_WORDS; ++i ){
      XStr_concat_Char( context->minibuffer_str, asdfjkl[i] );
      XStr_concat_Char( context->minibuffer_str, ':' );
      SDLSKK_XStr_concat_Str( context->minibuffer_str,
			      SDLSKK_DictItem_get_current( &item ) );
      XStr_concat_Char( context->minibuffer_str, ' ' );
      if( SDLSKK_DictItem_is_end( &item ) )
	break;
      
      SDLSKK_DictItem_next( &item );
    }
			      
    return context->minibuffer_str->buf;
  }
  
  return NULL;
}

char* SDLSKK_Context_get_display_str( SDLSKK_Context* context, char* cstr,
				      size_t size )
{
  SDLSKK_Char* str;
  str = SDLSKK_Context_get_display_Str( context );
  return SDLSKK_Str_to_cstr( str, cstr, size );
}
  
char* SDLSKK_Context_get_str( SDLSKK_Context* context, char* cstr,
			      size_t size )
{
  return SDLSKK_Str_to_cstr( context->str->buf, cstr, size );
}

int SDLSKK_Context_get_basic_mode( SDLSKK_Context* context )
{
  switch( context->mode ){
  case SDLSKK_J_KAKUTEI_MODE:
    return ( context->prefix->len == 0 );
  case SDLSKK_J_HENKAN_ON_MODE:
  case SDLSKK_J_HENKAN_OKURI_MODE:
  case SDLSKK_J_HENKAN_ACTIVE_MODE:
  case SDLSKK_ABBREV_MODE:
    return 0;
  case SDLSKK_LATIN_MODE:
  case SDLSKK_JISX0208_LATIN_MODE:
    return 1;
  default:
    return 0;
  }
}

void SDLSKK_Context_clear( SDLSKK_Context* context )
{
  context->mode = SDLSKK_J_KAKUTEI_MODE;
  SDLSKK_XStr_clear( context->str );
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  if( context->dict_item != NULL ){
    SDLSKK_DictItem_delete( context->dict_item );
    context->dict_item = NULL;
  }      
  SDLSKK_XStr_clear( context->display_str );
  SDLSKK_XStr_clear( context->minibuffer_str );
  context->cursor_pos = 0;
  context->henkan_buf_pos = 0;
  context->katakana_on = 0;
  if( context->child_context != NULL ){
    SDLSKK_Context_delete( context->child_context );
    context->child_context = NULL;
  }
  context->parent_context = NULL;
  SDLSKK_XStr_clear( context->cut_buffer );
  context->marked_point = NOT_MARKED;
}

int SDLSKK_Context_clear_text( SDLSKK_Context* context )
{
  SDLSKK_Mode old_mode;
  SDLSKK_XStr* old_cut_buffer;
  
  if( !SDLSKK_Context_get_basic_mode( context ) )
    return 0;

  old_mode = context->mode;
  old_cut_buffer = SDLSKK_XStr_new1( context->cut_buffer->buf );
  
  SDLSKK_Context_clear( context );
  
  context->mode = old_mode;
  SDLSKK_XStr_copy_Str( context->cut_buffer, old_cut_buffer->buf );
  SDLSKK_XStr_delete( old_cut_buffer );
  
  return 1;
}
