#include "sdlskk_internal.h"
#include <SDL.h>
#include <ctype.h>

struct KeyStrTable {
  char str[15];
  SDLKey sym;
  Uint16 unicode;
};

static struct KeyStrTable key_str_tbl[] = {
  { "SPC", SDLK_SPACE, ' ' },
  { "BS", SDLK_BACKSPACE, 0 },
  { "TAB", SDLK_TAB, '\t' },
  { "DEL", SDLK_DELETE, 0 },
  { "RET", SDLK_RETURN, 0 },
  { "UP", SDLK_UP, 0 },
  { "DOWN", SDLK_DOWN, 0 },
  { "RIGHT", SDLK_RIGHT, 0 },
  { "LEFT", SDLK_LEFT, 0 },
  { "INSERT", SDLK_INSERT, 0 },
  { "HOME", SDLK_HOME, 0 },
  { "END", SDLK_END, 0 },
  { "PAGEUP", SDLK_PAGEUP, 0 },
  { "PAGEDOWN", SDLK_PAGEDOWN, 0 },
  { "F1", SDLK_F1, 0 },
  { "F2", SDLK_F2, 0 },
  { "F3", SDLK_F3, 0 },
  { "F4", SDLK_F4, 0 },
  { "F5", SDLK_F5, 0 },
  { "F6", SDLK_F6, 0 },
  { "F7", SDLK_F7, 0 },
  { "F8", SDLK_F8, 0 },
  { "F9", SDLK_F9, 0 },
  { "F10", SDLK_F10, 0 },
  { "F11", SDLK_F11, 0 },
  { "F12", SDLK_F12, 0 },
  { "F13", SDLK_F13, 0 },
  { "F14", SDLK_F14, 0 },
  { "F15", SDLK_F15, 0 },
  { "HELP", SDLK_HELP, 0 },
  { "", 0, 0 },
};

SDL_keysym SDLSKK_convert_key_str( char* key )
{
  SDL_keysym result = {0, 0, 0, 0};
  int i;
  
  if( strlen(key) == 0 )
    return result; /* error */

  if( strlen(key) == 1 ){
    if( key[0] == ' ' ){
      result.unicode = ' ';
      result.sym = SDLK_SPACE;
      return result;
    }
    if( isgraph( key[0] ) ){
      result.unicode = key[0];
      return result;
    }
    return result; /* error */
  }

  /* strlen(key) >= 2 */
  if( key[1] == '-' ){
    result = SDLSKK_convert_key_str( &(key[2]) );
    result.unicode = tolower( result.unicode );
    switch( key[0] ){
    case 'C':
      result.mod |= KMOD_CTRL;
      return result;
    case 'M':
      result.mod |= (KMOD_ALT|KMOD_META);
      return result;
    default:
      return result; /* error */
    }
  }

  for( i=0; strcmp( key_str_tbl[i].str, "" ) != 0; ++i ){
    if( strcmp( key_str_tbl[i].str, key ) == 0 ){
      result.unicode = key_str_tbl[i].unicode;
      result.sym = key_str_tbl[i].sym;
      return result;
    }
  }
  
  return result; /* error */
}

SDLSKK_Keybind* SDLSKK_Keybind_new(void)
{
  SDLSKK_Keybind* keybind;
  
  keybind = SDLSKK_malloc( sizeof( SDLSKK_Keybind ) );
  keybind->top = NULL;
  return keybind;
}

void SDLSKK_Keybind_delete( SDLSKK_Keybind* keybind )
{
  SDLSKK_KeybindElement* element;
  SDLSKK_KeybindElement* next_element;

  for( element = keybind->top; element != NULL; ){
    next_element = element->next;
    free( element );
    element = next_element;
  }
  free( keybind );
  return;
}
    
static SDLSKK_KeybindElement*
Keybind_search_keybind_table( SDLSKK_Keybind* keybind,
			      SDL_keysym key )
{
  SDLSKK_KeybindElement* element;
  int ctrl, meta;
  Uint16 unicode;
  
  ctrl = (key.mod & KMOD_CTRL)?1:0 ;
  meta = (key.mod & (KMOD_ALT|KMOD_META))?1:0 ;
  if( iscntrl( key.unicode ) ){
    unicode = key.sym;
  }else{
    unicode = key.unicode;
  }

  for( element = keybind->top; element != NULL; element = element->next ){
    if( ctrl != ((element->key.mod & KMOD_CTRL)?1:0) )
      continue;
    if( meta != ((element->key.mod & (KMOD_ALT|KMOD_META))?1:0) )
      continue;
    if( unicode != 0 && unicode == element->key.unicode )
      return element;
    if( key.sym != 0 && key.sym == element->key.sym )
      return element;
  }
  
  return NULL;
}

SDLSKK_Command SDLSKK_Keybind_get_command( SDLSKK_Keybind* keybind,
					   SDL_keysym key, SDLSKK_Mode mode )
{
  SDLSKK_KeybindElement* bind;

  bind = Keybind_search_keybind_table( keybind, key );
  if( bind == NULL )
    return NULL;

  return bind->commands[mode];
}
  
void SDLSKK_Keybind_set_key( SDLSKK_Keybind* keybind, char* key_str,
			     char* command_str )
{
  SDLSKK_KeybindElement* bind;
  SDL_keysym key;
  SDLSKK_Command* command;
  
  key = SDLSKK_convert_key_str( key_str );
  command = SDLSKK_convert_command_str( command_str );

  if( key.unicode == 0 && key.sym == 0 )
    return;

  if( command == NULL )
    return;

  bind = Keybind_search_keybind_table( keybind, key );
  if( bind == NULL ){
    bind = SDLSKK_malloc( sizeof(SDLSKK_KeybindElement) );
    bind->key = key;
    bind->next = keybind->top;
    keybind->top = bind;
  }

  bind->commands = command;
  return;
}

void SDLSKK_Keybind_unset_key( SDLSKK_Keybind* keybind, char* key_str )
{
  SDLSKK_Keybind_set_key( keybind, key_str, "do-nothing" );
}

void SDLSKK_Keybind_set_default_key( SDLSKK_Keybind* keybind )
{
  char c;
  char str[3] = "a" ;
  
  SDLSKK_Keybind_set_key( keybind, "C-b", "backward-char" );
  SDLSKK_Keybind_set_key( keybind, "C-h", "backward-delete-char" );
  SDLSKK_Keybind_set_key( keybind, "C-d", "delete-char" );
  SDLSKK_Keybind_set_key( keybind, "C-f", "forward-char" );
  SDLSKK_Keybind_set_key( keybind, "C-j", "kakutei" );
  SDLSKK_Keybind_set_key( keybind, "C-m", "kettei" );
  SDLSKK_Keybind_set_key( keybind, " ", "space" );
  SDLSKK_Keybind_set_key( keybind, "C-g", "keyboard-quit" );
  SDLSKK_Keybind_set_key( keybind, "C-w", "kill-region" );
  SDLSKK_Keybind_set_key( keybind, "C-y", "yank" );
  SDLSKK_Keybind_set_key( keybind, "C-a", "beginning-of-line" );
  SDLSKK_Keybind_set_key( keybind, "C-e", "end-of-line" );
  SDLSKK_Keybind_set_key( keybind, "C-SPC", "set-mark-command" );
  SDLSKK_Keybind_set_key( keybind, "M-w", "copy" );
  SDLSKK_Keybind_set_key( keybind, "RIGHT", "forward-char" );
  SDLSKK_Keybind_set_key( keybind, "LEFT", "backward-char" );
  SDLSKK_Keybind_set_key( keybind, "DEL", "delete-char" );
  SDLSKK_Keybind_set_key( keybind, "BS", "backward-delete-char" );
  SDLSKK_Keybind_set_key( keybind, "RET", "kettei" );
  
  for( c='!'; c <= '~'; ++c ){
    str[0] = c;
    SDLSKK_Keybind_set_key( keybind, str, "graph-char" );
  }
  for( c='A'; c <= 'Z'; ++c ){
    str[0] = c;
    SDLSKK_Keybind_set_key( keybind, str, "upper-char" );
  }
  for( c='a'; c <= 'z'; ++c ){
    str[0] = c;
    SDLSKK_Keybind_set_key( keybind, str, "lower-char" );
  }

  SDLSKK_Keybind_set_key( keybind, "/", "abbrev-input" );
  SDLSKK_Keybind_set_key( keybind, "l", "latin-mode" );
  SDLSKK_Keybind_set_key( keybind, "x", "previous-candidate" );
  SDLSKK_Keybind_set_key( keybind, "L", "jisx0208-mode" );
  SDLSKK_Keybind_set_key( keybind, "q", "toggle-kana" );
}

